{- 2010-06-09 continuous normalization using the partiality monad
   Nisse asked me whether sized types helped here. 
   this is similar to "6.2 Continuous normalization" in my thesis (p. 131)
 -}

data Nat : Set
{ zero : Nat
; succ : Nat -> Nat
}

-- de Bruijn Terms

data Tm : Nat -> Set
{ vz  : [n : Nat] -> Tm (succ n)
; wk  : [n : Nat] -> Tm n -> Tm (succ n)
; lam : [n : Nat] -> Tm (succ n) -> Tm n
; app : [n : Nat] -> Tm n -> Tm n -> Tm n
; fix : [n : Nat] -> Tm n -> Tm n -> Tm n  -- CBV:  fix t v = t (fix t) v
}

-- environments and values (a bit ugly since we do not have mutual data)

data Env : Nat -> Set
{ empty : Env zero
; ext   : [n : Nat] -> Env n ->        
          [m : Nat] -> Tm (succ m) -> Env m -> -- a lambda-closure (unwinded) 
          Env (succ n)
} 

data Value : Set
{ clos : [m : Nat] -> (body : Tm (succ m)) -> (environ : Env m) -> Value
}

fun update : [n : Nat] -> Env n -> Value -> Env (succ n)
{ update n rho (clos m t rho') = ext n rho m t rho'
}

-- partiality monad

sized codata Lift [A : Set] : Size -> Set
{ now   : [i : Size] -> A        -> Lift A ($ i)
; later : [i : Size] -> Lift A i -> Lift A ($ i)
}

cofun bind : [A : Set] -> [B : Set] -> 
             [i : Size] -> Lift A i -> (A -> Lift B i) -> Lift B i
{ bind A B ($ i) (now   .A .i a) f = f a
; bind A B ($ i) (later .A .i x) f = later B i (bind A B i x f)
}

-- operational semantics
-- the evaluation function den and the application function vapp are defined mutually
-- by an outer corecursion into Lift and an inner recursion on Tm

mutual {

  cofun den : [i : Size] -> [n : Nat] -> Tm n -> Env n -> Lift Value i
  { den i .(succ n) (vz n)   (ext .n rho m t' rho') = now Value i (clos m t' rho')
  ; den i .(succ n) (wk n t) (ext .n rho m t' rho') = den i n t rho
  ; den i .n (lam n t) rho = now Value i (clos n t rho)
  ; den i .n (app n t1 t2) rho = 
     bind Value Value i (den i n t1 rho) (\ f ->
     bind Value Value i (den i n t2 rho) (\ v -> vapp i f v))
  ; den ($ i) .n (fix n t1 t2) rho = later Value i (den i n
     (app n (app n t1 (lam n (fix (succ n) (wk n t1) (vz n)))) t2) rho)
  }

  cofun vapp : [i : Size] -> Value -> Value -> Lift Value i
  { vapp ($ i) (clos n t rho) v = 
     later Value i (den i (succ n) t (update n rho v))
  } 

}

-- bisimulation of lifted values

sized codata LiftEq [A : Set] [+ Eq : A -> A -> Set] 
  : (i : Size) -> Lift A i -> Lift A i -> Set
{ eqnow   : [i : Size] -> [a : A] -> [a' : A] -> Eq a a' -> 
   LiftEq A Eq ($ i) (now A i a) (now A i a')
; eqlater : [i : Size] -> [x : Lift A i] -> [x' : Lift A i] ->
   LiftEq A Eq i x x' -> LiftEq A Eq ($ i) (later A i x) (later A i x')
-- missing the inductive constructors eqlaterl and eqlaterr
}
 
cofun bindcong : [i : Size] -> [A : Set] -> [B : Set] -> 
  [EqA : A -> A -> Set] -> [EqB : B -> B -> Set] -> 
  [x1 : Lift A i] -> [x2 : Lift A i] -> 
  [f1 : A -> Lift B i] -> [f2 : A -> Lift B i] ->
  LiftEq A EqA i x1 x2 ->
  ([v1 : A] -> [v2 : A] -> EqA v1 v2 -> LiftEq B EqB i (f1 v1) (f2 v2)) ->
  LiftEq B EqB i (bind A B i x1 f1) (bind A B i x2 f2)
{ bindcong ($ i) A B EqA EqB .(now A i v1) .(now A i v2) f1 f2
   (eqnow .A .EqA .i v1 v2 v1v2) f1f2 = f1f2 v1 v2 v1v2
; bindcong ($ i) A B EqA EqB .(later A i x1) .(later A i x2) f1 f2
   (eqlater .A .EqA .i x1 x2 x1x2) f1f2 = 
   eqlater B EqB i (bind A B i x1 f1) (bind A B i x2 f2) 
     (bindcong i A B EqA EqB x1 x2 f1 f2 x1x2 f1f2)
}

-- equality of values

{- cannot define EqV coinductively, because of negative occurrence 
sized codata EqV : Size -> Value -> Value -> Set
{ eqclos : [i : Size] -> 
   ((v1 : Value) -> (v2 : Value) -> EqV i v1 v2 -> EqV i ... ...) ->
   EqV ($ i) (clos m t rho) (m' t' rho')
}
-}

-- corecursion into Set, an unsafe feature for now
cofun EqV : Size -> Value -> Value -> Set 
{ EqV ($ i) (clos m1 t1 rho1) (clos m2 t2 rho2) = 
   [v1 : Value] -> [v2 : Value] -> EqV i v1 v2 ->
   LiftEq Value (EqV i) i
         (den i (succ m1) t1 (update m1 rho1 v1))
         (den i (succ m2) t2 (update m2 rho2 v2))
}

fun appcong : [i : Size] ->
  [f1 : Value] -> [f2 : Value] -> [v1 : Value] -> [v2 : Value] ->
  EqV i f1 f2 -> EqV i v1 v2 -> 
  LiftEq Value (EqV i) i (vapp i f1 v1) (vapp i f2 v2)
{ appcong ($ i) (clos m1 t1 rho1) (clos m2 t2 rho2) v1 v2 f1f2 v1v2 = 
    eqlater Value (EqV i) i
         (den i (succ m1) t1 (update m1 rho1 v1))
         (den i (succ m2) t2 (update m2 rho2 v2))  
         (f1f2 v1 v2 v1v2)
}

-- equality of environments

data EqEnv : [n : Nat] -> Env n -> Env n -> Set
{ eqempty : EqEnv zero empty empty
; eqext   : [n : Nat] -> 
            [rho1 : Env n] -> [rho2 : Env n] -> EqEnv n rho1 rho2 ->
            [m1 : Nat] -> [t1 : Tm (succ m1)] -> [eta1 : Env m1] ->
            [m2 : Nat] -> [t2 : Tm (succ m2)] -> [eta2 : Env m2] ->
            EqV # (clos m1 t1 eta1) (clos m2 t2 eta2) ->
            EqEnv (succ n) (ext n rho1 m1 t1 eta1) (ext n rho2 m2 t2 eta2)
}

fun equpdate : [n : Nat] -> 
               [rho1 : Env n] -> [rho2 : Env n] -> EqEnv n rho1 rho2 ->
               [v1 : Value] -> [v2 : Value] -> EqV # v1 v2 -> 
               EqEnv (succ n) (update n rho1 v1) (update n rho2 v2)
{ equpdate n rho1 rho2 rho1rho2 (clos m1 t1 eta1) (clos m2 t2 eta2) v1v2 =
    eqext n rho1 rho2 rho1rho2 m1 t1 eta1 m2 t2 eta2 v1v2
}

-- observational equality of terms

let EqTerm : (i : Size) -> (n : Nat) -> Tm n -> Tm n -> Set
  = \ i -> \ n -> \ t1 -> \ t2 -> 
    [rho1 : Env n] -> [rho2 : Env n] -> EqEnv n rho1 rho2 ->
    LiftEq Value (EqV #) i (den # n t1 rho1) (den # n t2 rho2)

-- reflexivity

cofun refl : [i : Size] -> [n : Nat] -> (t : Tm n) -> EqTerm i n t t
{ refl ($ i) .(succ n) (vz n) .(ext n rho1 m1 t1 eta1) .(ext n rho2 m2 t2 eta2)
   (eqext .n rho1 rho2 rho1rho2 m1 t1 eta1 m2 t2 eta2 v1v2) =
   eqnow Value (EqV #) i (clos m1 t1 eta1) (clos m2 t2 eta2) v1v2
; refl i .(succ n) (wk n t) .(ext n rho1 m1 t1 eta1) .(ext n rho2 m2 t2 eta2)
   (eqext .n rho1 rho2 rho1rho2 m1 t1 eta1 m2 t2 eta2 v1v2) =
   refl i n t rho1 rho2 rho1rho2
; refl ($ i) .n (lam n t) rho1 rho2 rho1rho2 = eqnow Value (EqV #) i
   (clos n t rho1) (clos n t rho2) 
   (\ v1 -> \ v2 -> \ v1v2 -> refl i (succ n) t 
     (update n rho1 v1) (update n rho2 v2) 
     (equpdate n rho1 rho2 rho1rho2 v1 v2 v1v2))
}